;;; nt-lfn.asm
;;; this is the 16 bit DOS TSR portion of the Long File Name API on Windows NT
;;; It will register the NT dll, hook Int 0x21, and then terminate and stay
;;; resident
;;; It then snoops int 21.  Any LFN calls will be handled here, and
;;; redirected to the DLL.  Some calls need special processing in
;;; both the DLL and here.

        
;;; this is a standard COM program.
[BITS 16]
[ORG 0x100]

;;; BOP Macros 
;;; from Microsoft Include File ISVBOP 
%define VDDRegister db 0xc4,0xc4,0x58,0x00
%define VDDDispatch db 0xc4,0xc4,0x58,0x02        
%define VDDUnregister db 0xc4,0xc4,0x58,0x01
        
        
        
[section .text]

;;; we need to check if we are installed or not ??  Maybe just call an LFN
;;; function and see if we get the unsupported code returned (0x7100 in AX with carry)
;;; OR - use the int 28? way of doing it.
        
        mov dx, versionString   ; dump some info out about what we are doing
        call printstr           ; 
        mov dx,start            ;
        call printstr           ;
        mov dx,dllPrintName     ;
        call printstr           ; 
        
        ;;  OK - register Our DLL - From VDD guide for Application Based Intercepts
        mov ax, 0               ; 
        mov si,dllName          ; si points to asciiz dll name
        mov di,dllInitializeName; di points to asciiz init routine
        mov bx, dllDispatchName ; bx points to asciiz dipatch routine
        ;; call Register Module
        VDDRegister
        jnc registerOK          ;
registerError:
        cmp ax, 1               ;
        je Lerror1              ;
        cmp ax, 2               ;
        je Lerror2              ;
        cmp ax, 3               ;
        je Lerror3              ;
        cmp ax, 4               ;
        je Lerror4              ;
        mov dx, errorUnknown    ;
        jmp endError            ;
Lerror1:
        mov dx, error1          ;
        jmp endError            ;
Lerror2:
        mov dx, dllDispatchPrintName
        call printstr
        mov dx, error2          ;
        jmp endError            ;
Lerror3:
        mov dx, dllInitializePrintName
        call printstr
        mov dx, error3          ;
        jmp endError            ;
Lerror4:
        mov dx, error4          ;
        jmp endError            ; 
endError:       
        call printstr           ;
        jmp doneTesting
registerOK:     
        mov [dllHandle],ax      ; store the handle away for safekeeping
        mov dx, success1        ; Tell Them We think it worked 
        call printstr           ;


;;; OK - Now time to get into the interrupt chain
        ;; now we need to hook DOS's own INT 21
        mov ah, 0x35            ; DOS Get Vector
        mov al, 0x21            ; DOS Handler
        int 0x21                ; invoke dos get vector
        mov [word origIRQVec],bx; store the old vector away for safekeeping
        mov [word origIRQVec+2],es; we will far jump to it later on
        
        ;; now we grab int 21 for ourselves
        mov ah, 0x25            ; DOS set vector
        mov al, 0x21            ; DOS Handler
        mov dx, interruptHandler; assume DS is OK
        int 0x21
        
        ;; Terminate and stay resident now
        mov dx, goodbye         ; say bye-bye
        call printstr
        mov ah, 0x31            ; TSR 
        mov al, 0               ; Return Code
        mov dx, 0x2000          ; num Paragraphs to reserve??? (check this XXX)
        int 0x21                ;

        
;;; OK Here is the Actual Hook routine that snoops on DOS calls and
;;; possibly redirects them into the NT DLL
interruptHandler:
        
        ;; nasm didn't like my mov ax, [sp] line so I set up a stack frame
        push bp
        mov bp,sp
        ;; use cs overrides for data access because thats all that we
        ;; know is ours
                
        ;; OK - save the return address to calleeCSIP
        ;; we are not going to do an iret, since we need to preserve CF
        ;; under NT Debug, this next line of code does not seem to work
        ;; if we are single stepping through things.  Nothing gets written to memory.
        mov [cs:saveAX], ax     ;
        mov ax, [bp+2]          ;
        mov [cs:word calleeCSIP],ax     ;
        mov ax, [bp+4]          ;
        mov [cs:word calleeCSIP +2],ax  ;
        ;; clean up the stack now
        pop bp                  ; close our stack frame
        ;; OK - We have saved the return address into our own variable
        ;; and everything else (except flags and ax ) is as it was when the int was generated
        
        ;; OK - Is it one of the Win95 LFN calls that we are interested in???
        ;;  if it is, and is a general LFN function that does not require
        ;; special handling then we jump to lfnFunc.  Special
        ;; routines will have special jump targets.
        mov ax, [cs:saveAX]     ; restore ax to original
        ;; OK - See if it is one of our functions  
        ;; we should probably keep track of what gets called the most
        ;; often and move those to the front (find first, generate short, ...)
              
        cmp ax, 0x5704          ; 100% (tested) Get Last Access Date and Time
        je lfnFunc              ; 
        cmp ax, 0x5705          ; 100%? (untested) Set Last Access Date and Time
        je lfnFunc              ;
        cmp ax, 0x5706          ; 100% (tested) Get Creation Date and Time
        je lfnFunc              ;
        cmp ax, 0x5707          ; 100%? (untested) Set Creation Date and Time
        je lfnFunc              ;
        cmp ax, 0x7139          ; 100% (tested) Make Directory
        je lfnFunc              ;
        cmp ax, 0x713A          ; 100% (tested) Remove Directory
        je lfnFunc              ;
        cmp ax, 0x713B          ; 100% (tested) Change Directory
        je near changeDir            ;
        cmp ah, 0x3B
        je near changeDir
        cmp ax, 0x7141          ; (Needs a lot of Work) Delete File
        je lfnFunc              ;
        cmp ax, 0x7143          ; (Partial) Get or set file attributes
        je lfnFunc              ;
        cmp ax, 0x7147          ; 100% (tested) Get Current Directory
        je lfnFunc              ;
        cmp ax, 0x714E          ; (Needs Work) Find First File
        je lfnFunc              ;
        cmp ax, 0x714F          ; (Needs Work) Find Next File
        je lfnFunc              ;
        cmp ax, 0x7156          ; 100% (tested) Rename File
        je lfnFunc              ;
        cmp ax, 0x7160          ; 2 subfuncs done and tested 1 subfunction needs to be emulated
        je lfnFunc              ;
        cmp ax, 0x716C          ; ~100% (Need Testing) Open/Create File
        je lfn716C              ; This needs a special 16 bit handler
        cmp ah, 0x3E            ; DOS File Close - must clean up our linked list too
        je near fileClose            ; 
        cmp ax, 0x71A0          ; 100% (tested) Get Volume Information
        je lfnFunc              ;
        cmp ax, 0x71A1          ; 100% (tested) Find Close
        je lfnFunc              ;
        cmp ax, 0x71A6          ; 100% (tested) Get File Info By Handle 
        je lfnFunc              ; 
        cmp ax, 0x71A7          ; (??Uncoded) FileTime/DOS Time Conversions
        je lfnFunc              ;
        cmp ax, 0x71A8          ; (Uncoded - partial) Generate Short Name - need to Emulate
        je lfnFunc              ;
        cmp ax, 0x71A9          ; (NEVER) Server Create - Do not need to emulate (hope)
        je lfnFunc              ; should letDosGetIt?
        cmp ax, 0x71AA          ; (NEVER) Subst - (3 functions) Do not need to Emulate (hope)
        je lfnFunc              ; shoule letDOSGetIt??       
        
        ;; doesn't look like its one of ours - call the DOS Handler
        jmp letDOSGetIt         ; this may need to be a jmp near ... if the stuff gets big

lfnFunc:        
        ;; wipe the return address and flags off the stack
        pop ax                  ; discard the flags and cs/ip
        pop ax                  ; we aren't going to do an IRET since that
        pop ax                  ; restores the flags, and we need to preserve carry
        mov ax, [cs:saveAX]     ; restore AX
        
        ;; we will pack the original value of ax into the high 16 bits
        ;;  the MS BOP mechanism wants AX for itself, but it appears
        ;;  that it preserves all 32 bits of EAX.
        shl eax, 16             
        mov ax, [cs:dllHandle]  ; load the bop handle into ax
        ;; we can probably lose this as we set it in the DLL automatically
        stc                     ; set carry just in case
        VDDDispatch             ; Call into the DLL now 
        jmp far [cs:calleeCSIP] ; simulate an IRET (we have already cleared stack)
lfn716C:
        ;; for file opens (and creates) the VDD must have a handle to the file
        ;; to do GetFileInfoByHandle, but we must have a dos handle as well
        ;; for the DOS read functions to work
        
        ;; create/open file
        mov ax, [cs:saveAX]
        ;; call LFN to get the short file name then call the non-LFN create/open
        ;; 6c00 is DOS 4 extended open/create
        shl eax, 16             ; put the original AX into the high bits of EAX
        mov ax, [cs:dllHandle]  ; put the BOP Handle into AX
        stc                     ; set Carry just in case (not needed ?)
        VDDDispatch             ; call the VDD
        ;; OK - see if carry is still set
        jnc keepOpening
        ;; problem - clean up and return, don't call dos at all
        shl eax, 16             ; not the best way of saving something but it works
        pop ax
        pop ax
        pop ax
        shr eax, 16             ; keep the error code from the VDD
        stc                     ; just in case??
        jmp far [cs:calleeCSIP] ; simulate an IRET now        
keepOpening:    
        ;; this should have converted DS:SI for us.  Now call old dos function

        ;; note - stack must not have the flags and return address popped at this point
        ;; we need to return to our own function, not the original one so we can trap
        ;; the handle
        push bp                 ; set up another stack frame until we figure out 
        mov bp,sp               ; how to use sp to address into the stack
        mov ax, cs              ; overwrite the Callee's CS:    IP with our own address 
        mov [bp+4], ax
        mov [bp+2], word returnLabel
        pop bp                  ; close the stack frame
        mov ax, 0x6c00          ; select DOS 4+ Create/Open File
        jmp far [cs:origIRQVec] ; Invoke DOS

returnLabel:                    ; DOS should return here now
        ;; now jump back to original program
        jc returnLeaveError
        mov bx, ax              ; move the handle to the Register the VDD wants
        mov ax, 1               ; select special Update File Handle VDD Function
        shl eax, 16             ; move to upper bits of EAX
        mov ax, [cs:dllHandle]  ; mov bop handle to AX
        VDDDispatch             ; call VDD
        mov ax, bx              ; put the handle back into AX now
        clc
        jmp far [cs:calleeCSIP] ; go back to user code
returnLeaveError:
        ;; This should not be happening.  Abort!!!
        ;; we have a special function in the dll to pop up a message and terminate
        mov ax, 0x3             ; special abort routine
        shl eax, 16
        mov ax, [cs:dllHandle]
        VDDDispatch             ;  I don't think will return
        stc                     ; but just in case
        jmp far [cs:calleeCSIP] ; simulate an IRET now


        
fileClose:
        ;; we must call into the vdd in case it opened the file to let it
        ;; close its own handle, then we can call dos to let it
        ;; close.   The VDD call can fail, but it should clear the carry
        ;; anyway.

        ;; note - stack must not have the flags and return address popped at this point
        ;; we need to return to our own function, not the original one so we can trap
        ;; the handle
        push bp                 ; set up another stack frame until we figure out 
        mov bp,sp               ; how to use sp to address into the stack
        mov ax, cs              ; overwrite the Callee's CS:    IP with our own address 
        mov [bp+4], ax
        mov [bp+2], word fileCloseReturnLabel
        pop bp                  ; close the stack frame
        mov ah, 0x3E            ; select DOS file close
        ;; store bx (we will use the ax storage spot, probably should rename
        mov [cs:saveAX],bx
        mov cx, 0
        jmp far [cs:origIRQVec] ; Invoke DOS

fileCloseReturnLabel:                    ; DOS should return here now
        ;; ok - did dos have an error
        jnc keepClosing
        mov cx, 0xABBA          ; signify that an error occurred to the VDD
keepClosing:    
        mov ax, 0x2
        shl eax, 16
        mov ax, [cs:dllHandle]
        mov bx, [cs:saveAX]
        VDDDispatch 
        jmp far [cs:calleeCSIP]

;;; OK - We will try changing directory in both NT and then call the non-LFN
;;; dos one as well
changeDir:        
        ;; we don't know if the short or long name version was called
        ;; so we will always set AX to the long version
        mov ax,0x713B
        shl eax, 16
        mov ax, [cs:dllHandle]
        VDDDispatch
        jnc keepChanging
        ;; well - we couldn't change - just return to the calling program
        pop ax
        pop ax
        pop ax
        stc
        jmp far [cs:calleeCSIP]
        
keepChanging:
        ;; now call the non-lfn dos change directory
        mov ah,0x3b
        jmp far [cs:origIRQVec]


        
letDOSGetIt:
        ;; just call dos, leave things on the stack
        ;; because dos will clean them up itself when it is done.
        mov ax, [cs:saveAX]     ; restore the AX register
        jmp far [cs: origIRQVec]; jump to DOS interrupt dispatch handler

;;; This is the end of the Interrupt Handler
       
        
         
        ;; in case we are already registered and have to abort
doneTesting:            
        ;; unregister the module now
        mov ax, [dllHandle]     ;
        ;; unregister module
        VDDUnregister
        mov dx, byeBye
        call printstr
        jmp theEnd              ; 
theEnd: 
        mov ax,0x4c00		; quit  
	int 0x21                ; invoke dos

;;; print out a $ terminated string by invoking DOS.
;;; dx holds pointer to string
;;; should make this a macro
printstr :     
	mov ah,9                ; select DOS print string function
	int 0x21                ; invoke dos
        ret
        

[section .data]
        ;; some general info string to print out
        versionString db 'TSR to Provide LFN support under Windows NT (12/29/97 1039)',13,10,'$';
        goodbye db 'Terminating And Staying Resident',13,10,'$'
        start  db 'Attempting to Register DLL Module', 13, 10, '$' ; [2]
        ;; error codes from Register Module
        error1 db 'DLL not Found',13,10,'$';
        error2 db 'Dispatch Routine Not Found',13,10,'$';
        error3 db 'Initialization Routine Not Found',13,10,'$';
        error4 db 'Out of Memory',13,10,'$';
        dispatchSuccess db 'Successful Dispatch',13,10,'$'; 
        dispatchError db 'Error when Dispatching',13,10,'$'; 
        errorUnknown db 'Unknown Error when attempting to Register Module',13,10,'$'; 
        success1 db 'Successful Registration!!!',13,10,'$';
        byeBye db 'Unregistered.  Time to go home',13,10,'$'; 

        ;; the name of our dll.  One format is for printing out
        dllName db 'nt-lfn.dll',0;
        dllPrintName db 'nt-lfn.dll',13,10,'$';

        ;; the entry point in the dll.  One format is for printing out
        dllDispatchName db 'NTLFNVDDDispatch',0;
        dllDispatchPrintName db 'NTLFNVDDDispatch',13,10,'$';

        ;; another dll function
        dllInitializeName db 'NTLFNVDDRegisterInit',0;
        dllInitializePrintName db 'NTLFNVDDRegisterInit',13,10,'$'; 

        ;; just initialize this to crap for now
        dllHandle dw 0xFFFF  ; stores the return value of Register Module, used by Dispatch 

        calleeCSIP dd 0            ; 
        saveAX dw   0              ; 
        origIRQVec dd 0            ; 













